/*
	File:		tilisten.c

	Contains:	TPI Module (that is actually in the Open Transport build) to
				gate connection indications to the listener one at a time.
				
				You may use this source code as a base for a similar stream module,
				but you may not name the module "tilisten" if you do so.

	Copyright:	 1994-1996 by Apple Computer, Inc., all rights reserved.

		
*/

#ifndef __OPENTPTMODULE__
#include <OpenTptModule.h>
#endif
#ifndef _MPS_STROPTS_
#include <stropts.h>
#endif
#ifndef _TIHDR_
#include <tihdr.h>
#endif

#ifndef qDebug
#define qDebug	0
#endif

/*
 * Forward function declarations
 */
 
int		tilisten_close(queue_t*, int, cred_t*);
int		tilisten_open(queue_t*, dev_t*, int, int, cred_t*);
int		tilisten_rput(queue_t*, mblk_t*);
int		tilisten_wput(queue_t*, mblk_t*);

/*
 * Module information for the streamtab structure
 */
 
static struct module_info minfo =  
{
	9003,				/* Module Number			*/
	"tilisten",			/* Name of module			*/
	0,					/* Minimum data size		*/
	INFPSZ,				/* Maximum data size		*/
	65536,				/* Hi water mark for queue	*/
	32768				/* Lo water mark for queue	*/
};

/*
 * Info for the "read" queue.
 */
 
static struct qinit rinit = 
{
	tilisten_rput,		/* Put routine for "incoming" data			*/
	0,					/* Service routine for "incoming" data		*/
	tilisten_open,		/* Our open routine							*/
	tilisten_close, 	/* Our close routine						*/
	0,					/* Our admin routine						*/
	&minfo				/* Our module_info							*/
};

/*
 * Info for the "write" queue
 */
 
static struct qinit winit =
{
	tilisten_wput,		/* Put routine for client data					*/
	0,					/* Service routine for client data				*/
	0,					/* open routine goes in read-side structure 	*/
	0,					/* close routine goes in read-side structure	*/
	0,					/* admin routine goes in read-side structure	*/
	&minfo				/* Our module_info								*/
};

/*
 * The streamtab structure
 */
 
static struct streamtab theStreamTab = 
{
	&rinit,			/* Our read-side qinit structure	*/
	&winit,			/* Our write-side qinit structure	*/
	0,				/* Not a mux - no structure info	*/
	0				/* Not a mux - no structure info	*/
};

/*******************************************************************************
** Local information
********************************************************************************/

enum
{
	kStateIdle			= 0,
	kStateIndUp			= 1,	/* Listen has been sent up the stream	*/
	kStateDisconDown	= 2,	/* Client issued reject (SndDisconnect)	*/
	kStateAcceptDown	= 3		/* Client issued Accept					*/
};

struct tilisten_data
{
	int				fState;
	long			fSequence;	/* Sequence number of current transaction		*/
	OTList			fIndList;	/* List of indications waiting to go upstream	*/
};

typedef struct tilisten_data	tilisten_data;

static char*	gModuleList	= NULL;

/*******************************************************************************
** Open routine
********************************************************************************/

int tilisten_open(queue_t* rdq, dev_t* dev, int flag, int sflag, cred_t* creds)
{
#pragma unused (flag)
	int				err;
	tilisten_data*	myData;
	/*
	 * If we're being reopened - just return
	 */
	if ( rdq->q_ptr != NULL )
		return kOTNoError;
	/*
	 * Make sure we're being opened properly - We only allow
	 * a "module" open.
	 */
	if ( sflag != MODOPEN )
		return ENXIO;
	
	err = mi_open_comm(&gModuleList, sizeof(tilisten_data), rdq, dev, flag, sflag, creds);
	
	if ( err == 0 )
		OTMemzero(rdq->q_ptr, sizeof(tilisten_data));
	return err;
}

/*******************************************************************************
** Close routine
********************************************************************************/

int tilisten_close(queue_t* rdq, int flags, cred_t* credP)
{
	mi_close_comm(&gModuleList, rdq);
	return 0;
}

/*******************************************************************************
** DestroyAllIndications
********************************************************************************/

void DestroyAllIndications(queue_t* q)
{
	tilisten_data*	data = (tilisten_data*)q->q_ptr;
	OTLink*			temp;
	
	while ( (temp = OTRemoveFirst(&data->fIndList)) != NULL )
	{
		freemsg(OTGetLinkObject(temp, mblk_t, b_next));
	}
	data->fState = kStateIdle;
}

/*******************************************************************************
** SendNextIndication
********************************************************************************/

int SendNextIndication(queue_t* q)
{
	tilisten_data*	data = (tilisten_data*)q->q_ptr;
	OTLink*			temp = OTRemoveFirst(&data->fIndList);
	
	if ( temp == NULL )
	{
		data->fState = kStateIdle;
	}
	else
	{
		mblk_t* mp = OTGetLinkObject(temp, mblk_t, b_next);
		
		data->fState = kStateIndUp;
		data->fSequence = ((T_conn_ind*)mp->b_rptr)->SEQ_number;
		putnext(q, OTGetLinkObject(temp, mblk_t, b_next));
	}
	return 0;
}

/*******************************************************************************
** Write-side put routine
********************************************************************************/

int tilisten_wput(queue_t* q, mblk_t* mp)
{
	long			prim;
	tilisten_data*	data;
	
	do
	{
		if ( mp->b_datap->db_type == M_FLUSH )
		{
			if ( mp->b_rptr[0] & FLUSHR )
			{
				DestroyAllIndications(q);
				break;
			}
		}
		if ( mp->b_datap->db_type != M_PROTO &&
			 mp->b_datap->db_type != M_PCPROTO ||
			 MBLK_SIZE(mp) < sizeof(long) )
		{
			break;
		}
		data = (tilisten_data*)q->q_ptr;
		prim = ((union T_primitives*)mp->b_rptr)->type;
		switch ( prim )
		{
			/*
			 * A T_CONN_RES means the user accepted the connection request.  If our
			 * state is not kStateIndUp, then that means that we probably sent up
			 * a T_DISCON_IND as the user was accepting it.
			 */
			case T_CONN_RES:
			{
				if ( data->fState == kStateIdle ||
					 ((T_conn_res*)mp->b_rptr)->SEQ_number != data->fSequence )
				{
				#if qDebug
					DebugStr("\ptilisten - Got unexpected T_CONN_RES");
				#endif
					/*
					 * We're confused - just send it on
					 */
					break;
				}
				/*
				 * Change our state to kStateAcceptDown and send the sucker on to
				 * the transport.
				 */
				data->fState = kStateAcceptDown;
				break;
			}
			
			case T_DISCON_REQ:
			{
				OTLink*	temp;
				/*
				 * If we're idle, or this isn't the sequence number we're currently
				 * working on, we're confused, so just send this puppy on.
				 */
				if ( data->fState == kStateIdle ||
					 ((T_discon_req*)mp->b_rptr)->SEQ_number != data->fSequence )
				{
				#if qDebug
					DebugStr("\ptilisten - Got unexpected T_DISCON_REQ");
				#endif
					break;
				}
				/*
				 * Change our state to kStateDisconDown and send the sucker on to
				 * the transport.
				 */
				data->fState = kStateDisconDown;
				break;
			}
		}
	} while ( false );
	putnext(q, mp);
	return 0;
}

/*******************************************************************************
** Read-side put routine
********************************************************************************/

int tilisten_rput(queue_t* q, mblk_t* mp)
{
	long			prim;
	tilisten_data*	data;
	do
	{
		if ( mp->b_datap->db_type == M_FLUSH )
		{
			if ( mp->b_rptr[0] & FLUSHR )
			{
				DestroyAllIndications(q);
				break;
			}
		}
		if ( mp->b_datap->db_type != M_PROTO &&
			 mp->b_datap->db_type != M_PCPROTO ||
			 MBLK_SIZE(mp) < sizeof(long) )
		{
			break;
		}
		prim = ((union T_primitives*)mp->b_rptr)->type;
		data = (tilisten_data*)q->q_ptr;
		switch ( prim )
		{
			case T_CONN_IND:
			{
				/*
				 * We've got a connection indication.  Queue it up, and if we're idle,
				 * Call SendNextIndication to deal with it.
				 */
			#if qDebug
				if ( data->fState == kStateIdle && data->fIndList.fHead != NULL )
				{
					DebugStr("\ptilisten_rput - state is idle, but indications are queued");
				}
			#endif
				OTAddLast(&data->fIndList, (OTLink*)&mp->b_next);
				if ( data->fState == kStateIdle )
					return SendNextIndication(q);
				return 0;
			}
		
			case T_DISCON_IND:
			{
				OTLink*	temp;
				/*
				 * If we got a disconnect indication, and we're idle, or the sequence
				 * number is not for the currently outstanding connection, we have to
				 * see if it's for one of the guys in our list of indications.
				 */
				if ( data->fState == kStateIdle ||
					 ((T_discon_ind*)mp->b_rptr)->SEQ_number != data->fSequence )
				{
					OTLink*	prev = (OTLink*)&data->fIndList.fHead;
					OTLink* next;
					
					while ( (next = prev->fNext) != NULL )
					{
						mblk_t* mpList = OTGetLinkObject(next, mblk_t, b_next);
						/*
						 * If this is the guy being cancelled, rip him off the list
						 * and free the message. Note that we don't change state at
						 * all because we're still handling an outstanding one in 
						 * the "foreground".  Also note that we don't send the
						 * T_DISCON_IND up to the client, since the client never
						 * saw the original T_CONN_IND.
						 */
						if ( ((T_discon_ind*)mp->b_rptr)->SEQ_number ==
							 ((T_discon_ind*)mpList->b_rptr)->SEQ_number )
						{
							prev->fNext = next->fNext;
							freemsg(mpList);
							freemsg(mp);
							return 0;
						}
						prev = next;
					}
				#if qDebug
					DebugStr("\ptilisten - Got unexpected T_DISCON_IND");
				#endif
					break;
				}
				/*
				 * Send on the disconnect indication to abort the current one
				 */
				putnext(q, mp);
				/*
				 * Then send up the next indication
				 */
				return SendNextIndication(q);
			}
			
			case T_OK_ACK:
			case T_ERROR_ACK:
			{
				long	prim2 = ((T_ok_ack*)mp->b_rptr)->CORRECT_prim;
				switch ( prim2 )
				{
					case T_CONN_RES:
					{
						if ( data->fState != kStateAcceptDown )
						{
						#if qDebug
							DebugStr("\ptilisten - Got unexpected ACK for T_CONN_RES");
						#endif
							break;
						}
						/*
						 * Send the ack on
						 */
						putnext(q, mp);
						/*
						 * Then send up the next waiting indication
						 */
						return SendNextIndication(q);
					}
					
					case T_DISCON_REQ:
					{
						if ( data->fState != kStateDisconDown )
						{
						#if qDebug
							DebugStr("\ptilisten - Got unexpected ACK for T_CONN_RES");
						#endif
							break;
						}
						/*
						 * Send the ack on
						 */
						putnext(q, mp);
						/*
						 * Then send up the next waiting indication
						 */
						return SendNextIndication(q);
					}
					
					default:
						break;
				}
				break;
			}
		}
	} while ( false );
	putnext(q, mp);
	return 0;
}


/* ------------------------------------------------------------------------

	THESE FUNCTIONS AND DATA STRUCTURES ARE SPECIFIC TO THE MACINTOSH IMPLEMENTATION
	OF STREAMS. THESE ROUTINES WOULD NOT EXIST IN OTHER STREAMS ENVIRONMENTS.
	
	WHEN PORTING OTHER STREAMS CODE TO THE MACINTOSH, THESE ROUTINES AND DATA
	STRUCTURES MUST BE ADDED.

  ------------------------------------------------------------------------ */

/*******************************************************************************
** install_info structure, and the GetInstallInfo function
**
** This is the structure that tells Open Transport about your module. OT calls
** your "GetInstallInfo" function to get this information.
**
********************************************************************************/

static struct install_info theInstallInfo =
{
	&theStreamTab,			// Stream Tab pointer
	kOTModIsModule, 		// Tell OT that we are a module, not a driver
	SQLVL_QUEUEPAIR,		// Synchronization level
	0,						// Shared writer list buddy
	0,						// Open Transport use - always set to 0
	0						// Flag - always set to 0
};
	
install_info* GetOTInstallInfo()
{
	return &theInstallInfo;
}
